
namespace eval math {
    variable PI [expr {atan2(0,-1)}]
    variable epsilon 1e-11
}

proc math::ConvertLengthsUnitsToSI { system } {
    switch $system {
	"m" { return 1.0 }
	"mm" { return 1e-3 }
	"cm" { return 1e-2 }
    }
    error "Error, using an unknown magnitude for the length units"
}

proc math::ConvertUnitsToSI { system magnitude } {

    switch $system {
	"N-m" { return 1.0 }
	"N-mm" {
	    switch $magnitude {
		N { return 1.0 }
		m { return 1e-3 }
		N/m { return 1e3 }
		Nm { return 1e-3 }
		Pa { return 1e6 }
	    }
	}
	"Kp-cm" {
	    switch $magnitude {
		N { return 9.8 }
		m { return 1e-2 }
		N/m { return 980.0 }
		Nm { return 0.098 }
		Pa { return 98000.0 }
	    }
	}
	"kN-m-MPa" {
	    switch $magnitude {
		N { return 1.0e3 }
		m { return 1.0 }
		N/m { return 1e3 }
		Nm { return 1e3 }
		Pa { return 1e6 }
	    }
	}
    }
    error "Error, using an unknown magnitude for the problem units ($system)"
}

proc math::Normalizevector { vec } {
    variable epsilon

    set norm [expr {sqrt([lindex $vec 0]*[lindex $vec 0]+[lindex $vec 1]*[lindex $vec 1]+\
	[lindex $vec 2]*[lindex $vec 2])}]

    if { abs($norm) > $epsilon } {
	return [list [expr {[lindex $vec 0]/$norm}] [expr {[lindex $vec 1]/$norm}] \
	    [expr {[lindex $vec 2]/$norm}] ]
    }
    return $vec
}

proc math::VectorProd { vec1 vec2 } {

    return [list [expr {[lindex $vec1 1]*[lindex $vec2 2]-[lindex $vec2 1]*[lindex $vec1 2]}] \
	[expr {[lindex $vec1 2]*[lindex $vec2 0]-[lindex $vec2 2]*[lindex $vec1 0]}] \
	[expr {[lindex $vec1 0]*[lindex $vec2 1]-[lindex $vec2 0]*[lindex $vec1 1]}]]
}

proc math::DotProd { vec1 vec2 } {
    return [expr {[lindex $vec1 0]*[lindex $vec2 0]+ \
	    [lindex $vec1 1]*[lindex $vec2 1]+ \
	    [lindex $vec1 2]*[lindex $vec2 2]}]
}

# returns vec1-vec2
proc math::VectorDiff { vec1 vec2 } {
    return [list [expr {[lindex $vec1 0]-[lindex $vec2 0]}] \
	[expr {[lindex $vec1 1]-[lindex $vec2 1]}] \
	[expr {[lindex $vec1 2]-[lindex $vec2 2]}]]
}

proc math::module { vec1 } {
    return [expr {sqrt([lindex $vec1 0]*[lindex $vec1 0]+\
	[lindex $vec1 1]*[lindex $vec1 1]+\
	[lindex $vec1 2]*[lindex $vec1 2])}]
}

proc math::ScalarDotVector { sc vec } {
    return [list [expr {$sc*[lindex $vec 0]}] [expr {$sc*[lindex $vec 1]}] [expr {$sc*[lindex $vec 2]}]]
}

# from 1 to 3
proc math::val { mat row col } {
    return [lindex $mat [expr {$col-1+3*($row-1)}]]
}

proc math::RotMat2OpenGLmat { mat } {
    
    return [list \
	[lindex $mat 0] [lindex $mat 3] [lindex $mat 6] 0 \
	[lindex $mat 1] [lindex $mat 4] [lindex $mat 7] 0 \
	[lindex $mat 2] [lindex $mat 5] [lindex $mat 8] 0 \
	0                0               0       1.0]
}

# axes must be: x,y or z
proc math::SetAxeInMatrix { mat axe values } {
    set axe [lsearch "x y z" $axe]
    for { set i 0 } { $i < 3 } { incr i } {
	lset mat [expr {$axe+3*$i}] [lindex $values $i]
    }
    return $mat
}

proc math::GetAxeInMatrix { mat axe } {
    set axe [lsearch "x y z" $axe]
    set values ""
    for { set i 0 } { $i < 3 } { incr i } {
	lappend values [lindex $mat [expr {$axe+3*$i}]]
    }
    return $values
}


proc math::BeamTransfMatrix { elm nodesname } {
    variable epsilon

    upvar $nodesname nodes

    set iprima [Normalizevector [list \
	[expr {[lindex $nodes([lindex $elm 1]) 0]-[lindex $nodes([lindex $elm 0]) 0]}] \
	[expr {[lindex $nodes([lindex $elm 1]) 1]-[lindex $nodes([lindex $elm 0]) 1]}] \
	[expr {[lindex $nodes([lindex $elm 1]) 2]-[lindex $nodes([lindex $elm 0]) 2]}]]]

    if { abs([lindex $iprima 0]) > $epsilon || abs([lindex $iprima 1]) } {
	if { abs([lindex $iprima 0]) > $epsilon } {
	    set jprima [list [expr {-1*[lindex $iprima 1]/double([lindex $iprima 0])}] 1.0 0.0]
	} else {
	    set jprima [list 1.0 0.0 0.0]
	}
	set jprima [Normalizevector $jprima]
	set kprima [VectorProd $iprima $jprima]
	if { abs([lindex $kprima 2]) < $epsilon } {
	    set jprima [ScalarDotVector -1.0 $jprima]
	    set kprima [VectorProd $iprima $jprima]
	}
    } else {
	set jprima [list 1.0 0.0 0.0]
	set kprima [VectorProd $iprima $jprima]
    }
    return [list \
	[lindex $iprima 0] [lindex $jprima 0] [lindex $kprima 0] \
	[lindex $iprima 1] [lindex $jprima 1] [lindex $kprima 1] \
	[lindex $iprima 2] [lindex $jprima 2] [lindex $kprima 2]]
}

proc math::CorrectBeamTransfMatrix { elm nodesname mat } {
    variable epsilon

    upvar $nodesname nodes

    set xaxe [Normalizevector [GetAxeInMatrix $mat x]]
    
    set iprima [Normalizevector [list \
	[expr {[lindex $nodes([lindex $elm 1]) 0]-[lindex $nodes([lindex $elm 0]) 0]}] \
	[expr {[lindex $nodes([lindex $elm 1]) 1]-[lindex $nodes([lindex $elm 0]) 1]}] \
	[expr {[lindex $nodes([lindex $elm 1]) 2]-[lindex $nodes([lindex $elm 0]) 2]}]]]

    set dotprod [DotProd $xaxe $iprima]
    if { abs($dotprod)>=1.0-$epsilon } { return $mat }

    set changeZaxe 1
    if { [val $mat 3 3] > $epsilon } { set changeZaxe 0 }
    if { $dotprod < -$epsilon } {
	set mat [SetAxeInMatrix $mat x [ScalarDotVector -1.0 $xaxe]]
	if { $changeZaxe } {
	    set mat [SetAxeInMatrix $mat z [ScalarDotVector -1.0 [GetAxeInMatrix $mat z]]]
	} else {
	    set mat [SetAxeInMatrix $mat y [ScalarDotVector -1.0 [GetAxeInMatrix $mat y]]]
	}
	set dotprod [expr {$dotprod*-1.0}]
    }
    if { abs($dotprod)>=1.0-$epsilon } { return $mat }
    if { $dotprod <= $epsilon } { return $mat } ;# this is an error: x orthogonal to beam

    set jprima [Normalizevector [GetAxeInMatrix $mat y]]
    set kprima [Normalizevector [VectorProd $iprima $jprima]]
    set jprima [VectorProd $kprima $iprima]

    set mat [SetAxeInMatrix $mat x $iprima]
    set mat [SetAxeInMatrix $mat y $jprima]
    set mat [SetAxeInMatrix $mat z $kprima]

    return $mat
}

proc math::ShellTransfMatrix  { elm nodesname } {
    variable epsilon

    upvar $nodesname nodes

    set kprima [Normalizevector [VectorProd \
	[list \
	    [expr {[lindex $nodes([lindex $elm 1]) 0]-[lindex $nodes([lindex $elm 0]) 0]}] \
	    [expr {[lindex $nodes([lindex $elm 1]) 1]-[lindex $nodes([lindex $elm 0]) 1]}] \
	    [expr {[lindex $nodes([lindex $elm 1]) 2]-[lindex $nodes([lindex $elm 0]) 2]}]] \
	[list \
	    [expr {[lindex $nodes([lindex $elm 2]) 0]-[lindex $nodes([lindex $elm 0]) 0]}] \
	    [expr {[lindex $nodes([lindex $elm 2]) 1]-[lindex $nodes([lindex $elm 0]) 1]}] \
	    [expr {[lindex $nodes([lindex $elm 2]) 2]-[lindex $nodes([lindex $elm 0]) 2]}]] \
       ]]
    
    if { abs([lindex $kprima 0]) > $epsilon || abs([lindex $kprima 2]) > $epsilon } {
	set iprima [Normalizevector [VectorProd [list 0.0 1.0 0.0] $kprima]]
    } else {
	set iprima [Normalizevector [VectorProd [list 0.0 0.0 1.0] $kprima]]
    }
    set jprima [VectorProd $kprima $iprima]

    return [list \
	[lindex $iprima 0] [lindex $jprima 0] [lindex $kprima 0] \
	[lindex $iprima 1] [lindex $jprima 1] [lindex $kprima 1] \
	[lindex $iprima 2] [lindex $jprima 2] [lindex $kprima 2]]
}

proc math::MatriuPerVec { mat vec } {

    set n [llength $vec]
    set ret ""
    for { set i 0 } { $i < $n } { incr i } {
	set val 0.0
	for { set j 0 } { $j < $n } { incr j } {
	    set val [expr {$val+[lindex $mat [expr {$j+$i*$n}]]*[lindex $vec $j]}]
	}
	lappend ret $val
    }
    return $ret
}

proc math::MatriuTPerVec { mat vec } {

    set n [llength $vec]
    set ret ""
    for { set i 0 } { $i < $n } { incr i } {
	set val 0.0
	for { set j 0 } { $j < $n } { incr j } {
	    set val [expr {$val+[lindex $mat [expr {$i+$j*$n}]]*[lindex $vec $j]}]
	}
	lappend ret $val
    }
    return $ret
}

proc math::GiveTriangleProjectedNatCoord { elm nodesname trielmsname pnt } {
    upvar $nodesname nodes
    upvar $trielmsname trielms

    set mat [ShellTransfMatrix [lrange $trielms($elm) 0 2] nodes]
    set tmppnt [MatriuTPerVec $mat $pnt]
    for { set i 0 } { $i < 3 } { incr i } {
	set tmpnodes($i) [MatriuTPerVec $mat $nodes([lindex $trielms($elm) $i])]
    }
    
    set retheight [expr {abs([lindex $tmppnt 2]-[lindex $tmpnodes(0) 2])}]
      
    set ax [expr {[lindex $tmpnodes(1) 0]-[lindex $tmpnodes(0) 0]}]
    set ay [expr {[lindex $tmpnodes(1) 1]-[lindex $tmpnodes(0) 1]}]
    
    set bx [expr {[lindex $tmpnodes(2) 0]-[lindex $tmpnodes(0) 0]}]
    set by [expr {[lindex $tmpnodes(2) 1]-[lindex $tmpnodes(0) 1]}]
    
    set cx [expr {[lindex $tmppnt 0]-[lindex $tmpnodes(0) 0]}]
    set cy [expr {[lindex $tmppnt 1]-[lindex $tmpnodes(0) 1]}]
    
    if { abs($ax) < $math::epsilon } {
	set N2 [expr {$cx/double($bx)}]
	set N1 [expr {($cy-$by*$N2)/double($ay)}]
    } else {
	set N2 [expr {($cy-$cx*$ay/double($ax))/double($by-$bx*$ay/double($ax))}]
	set N1 [expr {($cx-$bx*$N2)/double($ax)}]
    }
    return [list [expr {1.0-$N1-$N2}] $N1 $N2 $retheight]
}

proc math::CalcValueInTrianglePoint { elm nodesname trielmsname pnt gaussvalues } {
    variable LUmatTri6nodes

    upvar $nodesname nodes
    upvar $trielmsname trielms

    if { ![info exists LUmatTri6nodes] } {
	set LUmatTri6nodes [LU [list 0.5 0.5 0.0 \
		                     0.0 0.5 0.5 \
		                     0.5 0.0 0.5]]
    }
    foreach "N(0) N(1) N(2) H" [GiveTriangleProjectedNatCoord $elm nodes trielms $pnt] break
    for { set i 0 } { $i < 2 } { incr i } {
	if { $N($i) < -.01 || $N($i) > 1.01 } {
	    error "error N($i)=$N($i) 0 <= $i <= 2"
	}
    }
    set nodalvalues [LULinSolve $LUmatTri6nodes $gaussvalues]
    set val 0
    for { set i 0 } { $i < 3 } { incr i } {
	set val [expr {$val+$N($i)*[lindex $nodalvalues $i]}]
    }
    return $val
}

proc math::ElementCenter { elm nodesname } {
    upvar $nodesname nodes

    set center [list 0.0 0.0 0.0]
    set len [llength $elm]
    for { set i 0 } { $i < $len } { incr i } {
	lset center 0 [expr {[lindex $center 0]+[lindex $nodes([lindex $elm $i]) 0]}]
	lset center 1 [expr {[lindex $center 1]+[lindex $nodes([lindex $elm $i]) 1]}]
	lset center 2 [expr {[lindex $center 2]+[lindex $nodes([lindex $elm $i]) 2]}]
    }
    lset center 0 [expr {[lindex $center 0]/double($len)}]
    lset center 1 [expr {[lindex $center 1]/double($len)}]
    lset center 2 [expr {[lindex $center 2]/double($len)}]
    return $center
}

proc math::LU { mat } {
    # Descomposicin de Crout M=L*U de una matriz cuadrada
    # L matriz triangular inferior, U matriz triangular superior (con unos en la diagonal)
    # se retornan L y U guardadas en una matriz nica 
    # (la diagonal de U se sobreescribe, pero se sabe que es: 1.0,1.0,...,1.0)

    set n [expr {int(sqrt([llength $mat]))}]
    set LUmat [lrepeat 0.0 [expr {$n*$n}]]

    proc idx { r c n } { return [expr {($r-1)*$n+$c-1}] }

    set LU11 [expr {double([lindex $mat [idx 1 1 $n]])}]
    lset LUmat [idx 1 1 $n] $LU11 ;# L(1,1)=m(1,1); U(1,1)=1.0;
    for { set k 2 } { $k <= $n } { incr k } {
	lset LUmat [idx 1 $k $n] [expr {[lindex $mat [idx 1 $k $n]]/$LU11}]

	for { set i 2 } { $i < $k } { incr i } {
	    set suma 0.0
	    for { set j 1 } { $j < $i } { incr j } {
		set suma [expr {$suma+[lindex $LUmat [idx $i $j $n]]*[lindex $LUmat [idx $j $k $n]]}]
	    }
	    set LUii [expr {double([lindex $LUmat [idx $i $i $n]])}]
	    lset LUmat [idx $i $k $n] [expr {([lindex $mat [idx $i $k $n]]-$suma)/$LUii}]
	}
	lset LUmat [idx $k 1 $n] [lindex $mat [idx $k 1 $n]]

	for { set i 2 } { $i < $k } { incr i } {
	    set suma 0.0
	    for { set j 1 } { $j < $i } { incr j } {
		set suma [expr {$suma+[lindex $LUmat [idx $j $i $n]]*[lindex $LUmat [idx $k $j $n]]}]
	    }
	    lset LUmat [idx $k $i $n] [expr {[lindex $mat [idx $k $i $n]]-$suma}]
	}

	set suma 0.0
	for { set j 1 } { $j < $k } { incr j } {
	    set suma [expr {$suma+[lindex $LUmat [idx $k $j $n]]*[lindex $LUmat [idx $j $k $n]]}]
	}
	lset LUmat [idx $k $k $n] [expr {[lindex $mat [idx $k $k $n]]-$suma}]
    }
    return $LUmat
}

proc math::LULinSolve { LUmat b } {
    # solve the linear system (LU)*x=b     (LU= crout decomposition of M) 
    # returns x. on error, returns ""

    set n [expr {int(sqrt([llength $LUmat]))}]
    proc idx { r c n } { return [expr {($r-1)*$n+$c-1}] }

    set LU11 [expr {double([lindex $LUmat [idx 1 1 $n]])}]

    set eps 1e-15
    if { abs($LU11) < $eps } { error "error: cannot solve" }

    lset b 0 [expr {[lindex $b 0]/$LU11}]
    for { set i 2 } { $i <= $n } { incr i } {
	set LUii [expr {double([lindex $LUmat [idx $i $i $n]])}]
	if { abs($LUii) < $eps } { error "error: cannot solve" }
	set suma 0.0
	for { set j 1 } { $j < $i } { incr j } {
	    set suma [expr {$suma+[lindex $LUmat [idx $i $j $n]]*[lindex $b [expr {$j-1}]]}]
	}
	lset b [expr {$i-1}] [expr {([lindex $b [expr {$i-1}]]-$suma)/$LUii}]
    }
    # sustitucin adelante para resolver U*x=y ==>x
    for { set i [expr {$n-1}] } { $i >= 1 } { incr i -1 } {
	set suma 0.0
	for { set j [expr {$i+1}] } { $j <= $n } { incr j } {
	    set suma [expr {$suma+[lindex $LUmat [idx $i $j $n]]*[lindex $b [expr {$j-1}]]}]
	}
	lset b [expr {$i-1}] [expr {[lindex $b [expr {$i-1}]]-$suma}]
    }
    return $b
}


proc K {a b} {set a}
if {[string equal [info command lset] {}]} {
    proc lset {listName index val} {
	upvar $listName list
	set list [lreplace [K $list [set list {}]] $index $index $val]
    }
}

proc lrepeat { element count } {
    set retval ""
    for { set i 0 } { $i < $count } { incr i } {
	lappend retval $element
    }
    return $retval
}







